# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# ogbujipt.embedding.qdrant

'''
Vector databases embeddings using Qdrant
'''

import warnings
import itertools
# from typing import Sequence

# Qdrant install is optional for OgbujiPT
try:
    from qdrant_client import QdrantClient
    from qdrant_client.http import models
    QDRANT_AVAILABLE = True
except ImportError:
    QDRANT_AVAILABLE = False
    QdrantClient = object()  # Set up a dummy to satisfy the type hints
    models = None

# Option for running a Qdrant DB locally in memory
MEMORY_QDRANT_CONNECTION_PARAMS = {'location': ':memory:'}


class qdrant_collection:
    def __init__(self, name, embedding_model, db=None,
                 distance_function=None, **conn_params):
        '''
        Initialize a Qdrant client

        Args:
            name (str): of the collection

            embedding (SentenceTransformer): SentenceTransformer object of your choice
            https://huggingface.co/sentence-transformers

            
            db (optional QdrantClient): existing DB/client to use

            distance_function (str): Distance function by which vectors will be compared

            conn_params (mapping): keyword parameters for setting up QdrantClient
            See the main docstring (or run `help(QdrantClient)`)
            https://github.com/qdrant/qdrant-client/blob/master/qdrant_client/qdrant_client.py#L12

        Example:

        >>> from ogbujipt.text_helper import text_splitter
        >>> from ogbujipt.embedding_helper import qdrant_collection  # pip install qdrant_client
        >>> from sentence_transformers import SentenceTransformer  # pip install sentence_transformers
        >>> text = 'The quick brown fox\njumps over the lazy dog,\nthen hides under a log\nwith a frog.\n'
        >>> text += 'Should the hound wake up,\nall jumpers beware\nin a log, in a bog\nhe\'ll search everywhere.\n'
        >>> embedding_model = SentenceTransformer('all-MiniLM-L6-v2')
        >>> collection = qdrant_collection('my-text', embedding_model)
        >>> chunks = text_splitter(text, chunk_size=20, chunk_overlap=4, separator='\n')
        >>> collection.update(texts=chunks, metas=[{'seq-index': i} for (i, _) in enumerate(chunks)])
        >>> retval = collection.search('what does the fox say?', limit=1)
        retval
        '''
        self.name = name
        self.db = db
        # Check if the provided embedding model is a SentenceTransformer
        if embedding_model.__class__.__name__ == 'SentenceTransformer':
            self._embedding_model = embedding_model
        else:
            raise ValueError('embedding_model must be a SentenceTransformer object')
            
        if not self.db:
            if not QDRANT_AVAILABLE: 
                raise RuntimeError('Qdrant not installed, you can run `pip install qdrant-client`')

            # Create a Qdrant client
            if not conn_params:
                conn_params = MEMORY_QDRANT_CONNECTION_PARAMS
            self.db = QdrantClient(**conn_params)
        self._distance_function = distance_function or models.Distance.COSINE
        self._db_initialized = False

    def _first_update_prep(self, text):
        if text.__class__.__name__ != 'str':
            raise ValueError('text must be a string')

        # Make sure we have a vector size set; use a sample embedding if need be
        partial_embeddings = self._embedding_model.encode(text)
        self._vector_size = len(partial_embeddings)

        # Create a collection in the Qdrant client, and configure its vectors
        # Using REcreate_collection ensures overwrite for a clean, fresh, new collection
        self.db.recreate_collection(
            collection_name=self.name,
            vectors_config=models.VectorParams(
                size=self._vector_size, 
                distance=self._distance_function
                )
            )

        self._db_initialized = True

    def update(self, texts, metas=None):
        '''
        Update/insert into a Qdrant client's collection with the some chunks of text

        Args:
            texts (List[str]): Strings to be stored and indexed. For best results these should be of similar length.
            They'll be converted to embeddings for efficient lookup

            metas (List[dict]): Optional metadata per text, stored with the text and included whenever the text is
            retrieved via search/query
        '''
        if len(texts) == 0:
            warnings.warn('Empty sequence of texts provided. No action will be taken.')
            return

        if metas is None:
            metas = [{}]*(len(texts))
        else:
            if len(texts) > len(metas):
                warnings.warn(f'More texts ({len(texts)} provided than metadata {len(metas)}).'
                              'Extra metadata items will be ignored.')
                metas = itertools.chain(metas, [{}]*(len(texts)-len(metas)))
            elif len(metas) > len(texts):
                warnings.warn(f'Fewer texts ({len(texts)} provided than metadata {len(metas)}). '
                              'The extra text will be given empty metadata.')
                metas = itertools.islice(metas, len(texts))

        if not self._db_initialized:
            self._first_update_prep(texts[0])
            before_count = 0
        else:
            before_count = self.count()

        for ix, (text, meta) in enumerate(zip(texts, metas)):
            # Embeddings as float/vectors
            # The inline prints actually turn into a cool progress indicator in jupyter 😁
            embeddings = list(map(float, self._embedding_model.encode(text)))

            payload = dict(_text=text, **meta)

            # Upsert the embedded chunk and its payload into the collection
            self.db.upsert(
                collection_name=self.name,
                points=[
                    models.PointStruct(
                        id=ix + before_count,  # sequential IDs
                        vector=embeddings,
                        payload=payload
                        )
                    ]
                )
    
    def reset(self):
        '''
        Reset the Qdrant collection, deleting the collection and all its contents
        '''
        if not self._db_initialized:
            raise RuntimeError('Qdrant Collection must be initialized before deleting its contents.')
        
        self.db.delete_collection(collection_name=self.name)
        self._db_initialized = False

    def search(self, query, **kwargs):
        '''
        Perform a search on this Qdrant collection

        Args:
            query (str): string to compare against items in the collection

            kwargs: other args to be passed to qdrant_client.QdrantClient.search(). Common ones:
                    limit - maximum number of results to return (useful for top-k query)
        '''
        if not self._db_initialized:
            raise RuntimeError('Qdrant Collection must be initialized before searching its contents.')
        
        if query.__class__.__name__ != 'str':
            raise ValueError('query must be a string')
        embedded_query = self._embedding_model.encode(query)
        return self.db.search(collection_name=self.name, query_vector=embedded_query, **kwargs)

    def count(self):
        '''
        Return the count of items in this Qdrant collection
        '''
        if not self._db_initialized:
            raise RuntimeError('Qdrant Collection must be initialized before counting its contents.')
        # This ugly declaration just gets the count as an integer
        current_count = int(str(self.db.count(self.name)).partition('=')[-1])
        return current_count



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# ogbujipt.llm_wrapper
'''
Wrapper for LLMs, including via OpenAI API and locally hosted models

Side note: a lot of OpenAI tutorials suggest that you embed your
OpenAI private key into the code, which is a horrible, no-good idea

Extra reminder: If you set up your environment via .env file, make sure
it's in .gitignore or equivalent so it never gets accidentally committed!
'''

import os
import asyncio
import concurrent.futures
from functools import partial
from typing import List
import warnings

from amara3 import iri

from ogbujipt import config

try:
    from openai import OpenAI
    o = OpenAI(api_key='dummy')
    DEFAULT_OPENAI_API_BASE = o.base_url
    del o
except ImportError:
    OpenAI = None

# In many cases of self-hosted models you just get whatever model is loaded, rather than specifying it in the API
DUMMY_MODEL = 'DUMMY_MODEL'


# FIXME: Should probably be an ABC
class llm_wrapper:
    '''
    Base-level wrapper for LLMs
    '''
    def __init__(self, model=None, **kwargs):
        '''
        Args:
            model (str): Name of the model being wrapped

            kwargs (dict, optional): Extra parameters for the API, the model, etc.
        '''
        self.model = model
        self.parameters = kwargs


# Based on class BaseClient defn in https://github.com/openai/openai-python/blob/main/src/openai/_base_client.py
# plus class OpenAI in https://github.com/openai/openai-python/blob/main/src/openai/_client.py
# Omits lot-level HTTP stuff unless it turns out to be needed
OPENAI_KEY_ATTRIBS = ['api_key', 'base_url', 'organization', 'timeout', 'max_retries']


class openai_api(llm_wrapper):
    '''
    Wrapper for LLM hosted via OpenAI-compatible API (including OpenAI proper).
    Designed for models that provide simple completions from prompt.
    For chat-style models (including OpenAI's gpt-3.5-turbo & gpt-4), use openai_chat_api
            
    >>> from ogbujipt.llm_wrapper import openai_api
    >>> llm_api = openai_api(base_url='http://localhost:8000')
    >>> resp = llm_api('Knock knock!', max_tokens=128)
    >>> llm_api.first_choice_text(resp)
    '''
    def __init__(self, model=None, base_url=None, api_key=None, **kwargs):
        '''
        If using OpenAI proper, you can pass in an API key, otherwise environment variable
        OPENAI_API_KEY will be checked

        You can have multiple instances with different LLMs wrapped; perhaps one is OpenAI proper,
        another is a locally hosted model, etc. (a good way to save costs)

        Args:
            model (str, optional): Name of the model being wrapped. Useful for using
            OpenAI proper, or any endpoint that allows you to select a model

            base_url (str, optional): Base URL of the API endpoint

            api_key (str, optional): OpenAI API key to use for authentication

            kwargs (dict, optional): Extra parameters for the API or for the model host
        '''
        if OpenAI is None:
            raise ImportError('openai module not available; Perhaps try: `pip install openai`')
        if 'api_base' in kwargs:
            warnings.warn('api_base is deprecated; use base_url instead', DeprecationWarning, stacklevel=2)
            base_url = kwargs['api_base']
            del kwargs['api_base']
        if api_key is None:
            api_key = os.getenv('OPENAI_API_KEY', config.OPENAI_KEY_DUMMY)

        self.api_key = api_key
        self.parameters = config.attr_dict(kwargs)
        if base_url:
            # If the user includes the API version in the base, don't add it again
            scheme, authority, path, query, fragment = iri.split_uri_ref(base_url)
            path = path or kwargs.get('api_version', '/v1')
            self.base_url = iri.unsplit_uri_ref((scheme, authority, path, query, fragment))
        else:
            self.base_url = DEFAULT_OPENAI_API_BASE
        self.original_model = model or None
        self.model = model

    def __call__(self, prompt, api_func=None, **kwargs):
        '''
        Invoke the LLM with a completion request

        Args:
            prompt (str): Prompt to send to the LLM

            kwargs (dict, optional): Extra parameters to pass to the model via API

        Returns:
            dict: JSON response from the LLM
        '''
        # Have to build the upstream client object each time because it's not thread-safe
        # It seems to contain a thread lock object
        # Manifests in multiproc as TypeError: cannot pickle '_thread.RLock' object
        oai_client = OpenAI(api_key=self.api_key, base_url=self.base_url)
        api_func = api_func or oai_client.completions.create

        merged_kwargs = {**self.parameters, **kwargs}
        # XXX Figure this out more elegantly
        if 'model' in merged_kwargs:
            del merged_kwargs['model']

        result = api_func(model=self.model, prompt=prompt, **merged_kwargs)
        # print(result)
        if result.model == 'HOSTED_MODEL':
            result.model = self.hosted_model()
        return result

    def wrap_for_multiproc(self, prompt, **kwargs):
        '''
        Wrap the LLM invocation in an asyncio task

        Returns:
            asyncio.Task: Task for the LLM invocation
        '''
        merged_kwargs = {**self.parameters, **kwargs}
        # print(f'wrap_for_multiproc: {prompt =} {merged_kwargs =}')
        return asyncio.create_task(
            schedule_callable(self, prompt, **merged_kwargs))

    def hosted_model(self) -> str:
        '''
        Model introspection: Query the API to find what model is being run for LLM calls

        >>> from ogbujipt.llm_wrapper import openai_api
        >>> llm_api = openai_api(base_url='http://localhost:8000')
        >>> print(llm_api.hosted_model())
        '/models/TheBloke_WizardLM-13B-V1.0-Uncensored-GGML/wizardlm-13b-v1.0-uncensored.ggmlv3.q6_K.bin'
        '''
        if self.original_model:
            return self.original_model
        return self.available_models()[0]

    def available_models(self) -> List[str]:
        '''
        Query the API to find what model is being run for LLM calls

        Also includes model introspection, e.g.:

        >>> from ogbujipt.llm_wrapper import openai_api
        >>> llm_api = openai_api(base_url='http://localhost:8000')
        >>> print(llm_api.hosted_model())
        ['/models/TheBloke_WizardLM-13B-V1.0-Uncensored-GGML/wizardlm-13b-v1.0-uncensored.ggmlv3.q6_K.bin']
        '''
        try:
            import httpx  # noqa
        except ImportError:
            raise RuntimeError('Needs httpx installed. Try pip install httpx')

        resp = httpx.get(f'{self.base_url}/models').json()
        if 'data' not in resp:
            raise RuntimeError(f'Unexpected response from {self.base_url}/models:\n{repr(resp)}')
        return [ i['id'] for i in resp['data'] ]

    @staticmethod
    def first_choice_text(response):
        '''
        Given an OpenAI-compatible API simple completion response, return the first choice text
        '''
        try:
            return response.choices[0].text
        except AttributeError:
            raise RuntimeError(
                f'''Response does not appear to be an OpenAI API completion structure, as expected:
{repr(response)}''')


class openai_chat_api(openai_api):
    '''
    Wrapper for a chat-style LLM hosted via OpenAI-compatible API (including OpenAI proper).
    Supports local chat-style models as well as OpenAI's gpt-3.5-turbo & gpt-4

    You need to set an OpenAI API key in your environment, or pass it in, for this next example

        >>> from ogbujipt.llm_wrapper import openai_chat_api, prompt_to_chat
    >>> llm_api = openai_chat_api(model='gpt-3.5-turbo')
    >>> resp = llm_api(prompt_to_chat('Knock knock!'))
    >>> llm_api.first_choice_message(resp)
    '''
    def __call__(self, messages, api_func=None, **kwargs):
        '''
        Invoke the LLM with a completion request

        Args:
            messages (list): Series of messages representing the chat history
            e.f. messages=[{"role": "user", "content": "Hello world"}])

            kwargs (dict, optional): Extra parameters to pass to the model via API

        Returns:
            dict: JSON response from the LLM
        '''
        # Have to build the upstream client object each time because it's not thread-safe
        # It seems to contain a thread lock object
        # Manifests in multiproc as TypeError: cannot pickle '_thread.RLock' object
        oai_client = OpenAI(api_key=self.api_key, base_url=self.base_url)
        api_func = api_func or oai_client.chat.completions.create

        # XXX Figure this out more elegantly
        # if 'api_key' in merged_kwargs:
        #     del merged_kwargs['api_key']
        # if 'base_url' in merged_kwargs:
        #     del merged_kwargs['base_url']
        merged_kwargs = {**self.parameters, **kwargs}
        if 'model' in merged_kwargs:
            del merged_kwargs['model']

        result = api_func(model=self.model, messages=messages, **merged_kwargs)
        # print(result)
        if result.model == 'HOSTED_MODEL':
            result.model = self.hosted_model()
        return result

    @staticmethod
    def first_choice_message(response):
        '''
        Given an OpenAI-compatible API chat completion response, return the first choice message content
        '''
        try:
            return response.choices[0].message.content
        except AttributeError:
            raise RuntimeError(
                f'''Response does not appear to be an OpenAI API chat-style completion structure, as expected:
{repr(response)}''')


class ctransformer:
    '''
    ctransformers wrapper for LLMs
    '''
    def __init__(self, model=None, **kwargs):
        '''
        Args:
            model (XYZ): Name of the model being wrapped

            kwargs (dict, optional): Extra parameters for the API, the model, etc.

        # gpu_layers - # of layers to offload to GPU. Set to 0 if no GPU acceleration is available on your system
        llm = AutoModelForCausalLM.from_pretrained(
            'TheBloke/LlongOrca-13B-16K-GGUF', model_file='llongorca-13b-16k.q4_K_M.gguf',
            model_type='llama', gpu_layers=50)

        ctrans_wrapper = ctrans_wrapper(model=llm)

        Defined model params from https://github.com/marella/ctransformers/blob/main/ctransformers/llm.py

        top_k="The top-k value to use for sampling.",
        top_p="The top-p value to use for sampling.",
        temperature="The temperature to use for sampling.",
        repetition_penalty="The repetition penalty to use for sampling.",
        last_n_tokens="The number of last tokens to use for repetition penalty.",
        seed="The seed value to use for sampling tokens.",
        max_new_tokens="The maximum number of new tokens to generate.",
        stop="A list of sequences to stop generation when encountered.",

        Operational params:

        stream="Whether to stream the generated text.",
        reset="Whether to reset the model state before generating text.",
        batch_size="The batch size to use for evaluating tokens in a single prompt.",
        threads="The number of threads to use for evaluating tokens.",
        context_length="The maximum context length to use.",
        gpu_layers="The number of layers to run on GPU.",
        '''
        # if AutoModelForCausalLM is None:
        #     raise ImportError('ctransformers module not available; Perhaps try: `pip install ctransformers`')
        if model is None:
            raise ValueError('Must supply a model')
        self.model = model
        self.parameters = kwargs

    def __call__(self, prompt, **kwargs):
        '''
        Invoke the LLM with a completion request

        Args:
            prompt (str): Prompt to send to the LLM

            kwargs (dict, optional): Extra parameters to pass to the model via API

        Returns:
            dict: JSON response from the LLM
        '''
        # return self.model.generate(prompt, **kwargs)
        # yield from self.model.generate(prompt, **kwargs)
        return self.model(prompt, **kwargs)


def prompt_to_chat(prompt, system=None):
    '''
    Convert a prompt string to a chat-style message list

    Args:
        prompt (str): Prompt to convert

    Returns:
        list: single item with just the given prompt as user message for chat history
        e.g. messages=[{"role": "user", "content": "Hello world"}])
    '''
    # return [{'role': 'user', 'content': m} for m in prompt.split('\n')]
    messages = [] if system is None else [{'role': 'system', 'content': system}]
    messages.append({'role': 'user', 'content': prompt})
    return messages


async def schedule_callable(callable, *args, **kwargs):
    '''
    Schedule long-running/blocking function call in a separate process,
    wrapped to work well in an asyncio event loop

    Basically hides away a bunch of the multiprocessing webbing

    e.g. `llm_task = asyncio.create_task(schedule_callable(llm, prompt))`

    Can then use asyncio.wait(), asyncio.gather(), etc. with `llm_task`

    Args:
        callable (callable): Callable to be scheduled

    Returns:
        response: Response object
    '''
    # print(f'schedule_callable: {kwargs=}')
    # Link up the current async event loop for multiprocess execution
    loop = asyncio.get_running_loop()
    executor = concurrent.futures.ProcessPoolExecutor()
    # Need to partial execute to get in any kwargs for the target callable
    prepped_callable = partial(callable, **kwargs)
    # Spawn a separate process for the LLM call
    response = await loop.run_in_executor(executor, prepped_callable, *args)
    return response



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# ogbujipt/demo/qa_discord.py
'''
Advanced demo of a Discord chatbot with an LLM back end

Demonstrates async processing & Discord API integration

Note: Simple demo, which doesn't do any client-side job management,
so for example if a request is sent, and a second comes in before it has completed,
the LLM back end is relied on to cope.

Additional prerequisites: discord.py

Also need to make sure Python has root SSL certificates installed
On MacOS this is via double-clicking `Install Certificates.command`

Requires the following environment variables:

```env
DISCORD_TOKEN={your-bot-token}
LLM_BASE=http://my-llm-host:8000
LLM_TEMP=0.5
```

For some deeper discussion of setting up the environment: https://github.com/OoriData/OgbujiPT/discussions/36

To launch the bot:

```shell
python demo/qa_discord.py
```

You can then @mention the bot in a Discord channel where it's been added & chat with it

For hints on how to modify this to use OpenAI's actual services, see demo/alpaca_fix_xml.py
'''

import os

import discord

from ogbujipt.llm_wrapper import openai_chat_api, prompt_to_chat

# Enable all standard intents, plus message content
# The bot app you set up on Discord will require this intent (Bot tab)
intents = discord.Intents.default()
intents.message_content = True

client = discord.Client(intents=intents)


async def send_llm_msg(msg):
    '''
    Schedule the LLM request
    '''
    # See demo/alpaca_multitask_fix_xml.py for some important warnings here
    # oapi.parameters
    response = await oapi.wrap_for_multiproc(prompt_to_chat(msg), max_tokens=512)
    print(response)

    print('\nFull response data from LLM:\n', response)

    # Response is a json-like object; we just need the message text
    response_text = oapi.first_choice_message(response)
    print('\nResponse text from LLM:\n', response_text)

    return response_text


@client.event
async def on_message(message):
    # Ignore the bot's own messages & respond only to @mentions
    # The client.user.id check creens out @everyone & @here pings
    # FIXME: Better content check—what if the bot's id is a common word?
    if message.author == client.user \
            or not client.user.mentioned_in(message) \
            or str(client.user.id) not in message.content:
        return

    # Send throbber placeholder message to discord:
    return_msg = await message.channel.send('<a:oori_throbber:1142173241499197520>')

    # Assumes a single mention, for simplicity. If there are multiple,
    # All but the first will just be bundled over to the LLM
    mention_str = f'<@{client.user.id}>'
    clean_msg = message.content.partition(mention_str)
    clean_msg = clean_msg[0] + clean_msg[2]

    response = await send_llm_msg(clean_msg)

    await return_msg.edit(content=response[:2000])  # Discord messages cap at 2k characters


@client.event
async def on_ready():
    print(f"Bot is ready. Connected to {len(client.guilds)} guild(s).")


def main():
    # A real app would probably use a discord.py cog w/ this as data member
    global oapi

    DISCORD_TOKEN = os.getenv('DISCORD_TOKEN')
    assert DISCORD_TOKEN
    llm_base = os.getenv('LLM_BASE', 'http://localhost:8000')
    llmtemp = float(os.getenv('LLM_TEMP', '0.9'))

    # Set up API connector; OpenAI API emulation with supplied API base, fixed temp, from the environment
    oapi = openai_chat_api(base_url=llm_base, temperature=llmtemp)

    # launch Discord client event loop
    client.run(DISCORD_TOKEN)


if __name__ == '__main__':
    # Entry point protects against multiple launching of the overall program when a child process imports
    # viz https://docs.python.org/3/library/multiprocessing.html#multiprocessing-safe-main-import
    main()



'''
pytest test

or

pytest test/test_airoboros_obedient_prompting.py
'''
# import pytest

from ogbujipt.prompting.basic import format
from ogbujipt.prompting.model_style import AIROBOROS_OBEDIENT_DELIMITERS

def test_airoboros_obedient_inputs_prompt_building(PROMPTING_USER_QUERY, PROMPTING_CONTEXT_LIST):
    EXPECTED_PROMPT = 'BEGININPUT\nBEGINCONTEXT\nauthor: Isaac Newton\ntitle: Principia Mathematica Volume 1\ndate: 1910\nISBN-13: 978-0521626063\nENDCONTEXT\nAll mathematical propositions\' are of the form S is P\'; S\' is the subject, P\' the predicate.\nENDINPUT\nBEGININPUT\nBEGINCONTEXT\nauthor: Isaac Newton\ntitle: Principia Mathematica Volume 2\ndate: 1912\nISBN-13: 978-0521626070\nENDCONTEXT\nIt is desirable to find some convenient phrase which shall denote the whole collection of values of a function, or the whole collection of entities of a given type.\nENDINPUT\nBEGININPUT\nBEGINCONTEXT\nauthor: Isaac Newton\ntitle: Principia Mathematica Volume 3\ndate: 1913\nISBN-13: 978-0521626087\nENDCONTEXT\nIn this volume we shall define the arithmetical relations between finite cardinal numbers and also their addition and multiplication.\nENDINPUT\nBEGININPUT\nBEGINCONTEXT\nauthor: Isaac Newton\ntitle: Principia Mathematica Volume 4\ndate: 1915\nISBN-13: 978-6969696969\nENDCONTEXT\nFinally, we can get onto the real purpose of this work; to explain why 9 + 10 = 21\nENDINPUT\nBEGININSTRUCTION\nwhat\'s nine plus ten?\nENDINSTRUCTION\n'  # noqa: E501

    prompt = format(
        PROMPTING_USER_QUERY,
        contexts=PROMPTING_CONTEXT_LIST,
        delimiters=AIROBOROS_OBEDIENT_DELIMITERS,
        context_with_metadata=True,
        )
    assert prompt == EXPECTED_PROMPT


# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# ogbujipt/demo/chat_web_selects.py
'''
Advanced, "Chat my docs" demo, using docs from the web

Download one or more web pages and query an LLM using them as context.
Works especially well with airoboros self-hosted LLM.

Vector store: Qdrant - https://qdrant.tech/
    Alternatives: pgvector, Chroma, Faiss, Weaviate, etc.
Text to vector (embedding) model: 
    Alternatives: https://www.sbert.net/docs/pretrained_models.html / OpenAI ada002

Needs access to an OpenAI-like service. Default assumption is self-hosted
via e.g. llama-cpp-python or text-generation-webui

Assume for the following it's at host my-llm-host, port 8000

pip install prerequisites, in addition to OgbujiPT cloned dir:

click sentence_transformers qdrant-client httpx html2text amara3.xml

```sh
python demo/chat_web_selects.py --apibase http://my-llm-host:8000 "www.newworldencyclopedia.org/entry/Igbo_People"
```

An example question might be "Who are the neighbors of the Igbo people?"
'''
# en.wikipedia.org/wiki/Igbo_people|ahiajoku.igbonet.com/2000/|en.wikivoyage.org/wiki/Igbo_phrasebook"
import asyncio
import os

import click
import httpx
import html2text

from ogbujipt.llm_wrapper import openai_chat_api, prompt_to_chat
from ogbujipt.text_helper import text_splitter
from ogbujipt.embedding_helper import qdrant_collection


# Avoid re-entrace complaints from huggingface/tokenizers
os.environ['TOKENIZERS_PARALLELISM'] = 'false'

# default https://huggingface.co/sentence-transformers/all-MiniLM-L6-v2
DOC_EMBEDDINGS_LLM = 'all-MiniLM-L12-v2'

COLLECTION_NAME = 'chat-web-selects'
USER_PROMPT = 'What do you want to know from the site(s)?\n'

# Hard-code for demo
EMBED_CHUNK_SIZE = 200
EMBED_CHUNK_OVERLAP = 20
DOTS_SPACING = 0.2  # Number of seconds between each dot printed to console


async def indicate_progress(pause=DOTS_SPACING):
    while True:
        print('.', end='', flush=True)
        await asyncio.sleep(pause)


async def read_site(url, collection):
    # Crude check; good enough for demo
    if not url.startswith('http'): url = 'https://' + url  # noqa E701
    print('Downloading & processing', url)
    async with httpx.AsyncClient(verify=False) as client:
        resp = await client.get(url)
        html = resp.content.decode(resp.encoding or 'utf-8')

    text = html2text.html2text(html)

    # Split text into chunks
    chunks = text_splitter(text, chunk_size=EMBED_CHUNK_SIZE,
                           chunk_overlap=EMBED_CHUNK_OVERLAP, separator='\n')

    # print('\n\n'.join([ch[:100] for ch in chunks]))
    # Crude—for demo. Set URL metadata for all chunks to doc URL
    metas = [{'url': url}]*len(chunks)
    # Add the text to the collection
    collection.update(texts=chunks, metas=metas)
    print(f'{collection.count()} chunks added to collection')


async def async_main(oapi, sites):
    # Automatic download of embedding model from HuggingFace
    # Seem to be reentrancy issues with HuggingFace; defer import
    from sentence_transformers import SentenceTransformer
    embedding_model = SentenceTransformer(DOC_EMBEDDINGS_LLM)
    # Sites fuel in-memory Qdrant vector DB instance
    collection = qdrant_collection(COLLECTION_NAME, embedding_model)

    # Download & process sites in parallel, loading their content & metadata into a knowledgebase
    url_task_group = asyncio.gather(*[
        asyncio.create_task(read_site(site, collection)) for site in sites.split('|')])
    indicator_task = asyncio.create_task(indicate_progress())
    tasks = [indicator_task, url_task_group]
    done, _ = await asyncio.wait(
        tasks, return_when=asyncio.FIRST_COMPLETED)

    # Main chat loop
    done = False
    while not done:
        print()
        user_question = input(USER_PROMPT)
        if user_question.strip() == 'done':
            break

        docs = collection.search(user_question, limit=4)

        print(docs)
        if docs:
            # Collects "chunked_doc" into "gathered_chunks"
            gathered_chunks = '\n\n'.join(
                doc.payload['_text'] for doc in docs if doc.payload)

            # Build system message with the doc chunks as provided context
            # In practice we'd use word loom to load the propts, as demoed in multiprocess.py
            sys_prompt = '''\
You are a helpful assistant, who answers questions directly and as briefly as possible.
Consider the following context and answer the user\'s question.
If you cannot answer with the given context, just say so.\n\n'''
            sys_prompt += gathered_chunks + '\n\n'
            messages = prompt_to_chat(user_question, system=sys_prompt)

            print('-'*80, '\n', messages, '\n', '-'*80)

            # The rest is much like in demo/alpaca_multitask_fix_xml.py
            model_params = dict(
                max_tokens=1024,  # Limit number of generated tokens
                top_p=1,  # AKA nucleus sampling; can increase generated text diversity
                frequency_penalty=0,  # Favor more or less frequent tokens
                presence_penalty=1,  # Prefer new, previously unused tokens
                temperature=0.1
                )

            indicator_task = asyncio.create_task(indicate_progress())
            llm_task = oapi.wrap_for_multiproc(messages, **model_params)
            tasks = [indicator_task, llm_task]
            done, _ = await asyncio.wait(
                tasks, return_when=asyncio.FIRST_COMPLETED)

            # Instance of openai.openai_object.OpenAIObject, with lots of useful info
            retval = next(iter(done)).result()
            print(type(retval))
            # Response is a json-like object; extract the text
            print('\nFull response data from LLM:\n', retval)

            # response is a json-like object; 
            # just get back the text of the response
            response_text = oapi.first_choice_message(retval)
            print('\nResponse text from LLM:\n\n', response_text)


# Command line arguments defined in click decorators
@click.command()
@click.option('--apibase', default='http://127.0.0.1:8000', help='OpenAI API base URL')
@click.option('--openai-key',
              help='OpenAI API key. Leave blank to specify self-hosted model via --host & --port')
@click.option('--model', default='', type=str, 
              help='OpenAI model to use (see https://platform.openai.com/docs/models).'
              'Use only with --openai-key')
@click.argument('sites')
def main(apibase, openai_key, model, sites):
    # Use OpenAI API if specified, otherwise emulate with supplied URL info
    if openai_key:
        oapi = openai_chat_api(api_key=openai_key, model=(model or 'gpt-3.5-turbo'))
    else:
        oapi = openai_chat_api(model=model, base_url=apibase)

    asyncio.run(async_main(oapi, sites))


if __name__ == '__main__':
    main()



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# ogbujipt.embedding.pgvector

'''
Vector databases embeddings using PGVector
'''

# import warnings
# import itertools
import json
from typing   import Sequence
from uuid     import UUID
from datetime import datetime, timezone

# Handle key imports
try:
    import asyncpg
    from pgvector.asyncpg import register_vector
    PREREQS_AVAILABLE = True
except ImportError:
    PREREQS_AVAILABLE = False
    asyncpg = None
    register_vector = object()  # Set up a dummy to satisfy the type hints

# ======================================================================================================================
# PG only supports proper query arguments (e.g. $1, $2, etc.) for values, not for table or column names
# Table names are checked to be legit sequel table names, and embed_dimension is assured to be an integer
CREATE_VECTOR_EXTENSION = 'CREATE EXTENSION IF NOT EXISTS vector;'

CHECK_TABLE_EXISTS = '''-- Check if a table exists
SELECT EXISTS (
    SELECT FROM pg_tables
    WHERE tablename = $1
);
'''

# Generic SQL template for creating a table to hold embedded documents
CREATE_DOC_TABLE = '''-- Create a table to hold embedded documents
CREATE TABLE IF NOT EXISTS {table_name} (
    id BIGSERIAL PRIMARY KEY,
    embedding VECTOR({embed_dimension}),  -- embedding vectors (array dimension)
    content TEXT NOT NULL,                -- text content of the chunk
    title TEXT,                           -- title of file
    page_numbers INTEGER[],               -- page number of the document that the chunk is found in
    tags TEXT[]                           -- tags associated with the chunk
);
'''

INSERT_DOCS = '''-- Insert a document into a table
INSERT INTO {table_name} (
    embedding,
    content,
    title,
    page_numbers,
    tags
) VALUES ($1, $2, $3, $4, $5);
'''

QUERY_DOC_TABLE = '''-- Semantic search a document
SELECT
    (embedding <=> '{query_embedding}') AS cosine_similarity,
    title,
    content,
    page_numbers,
    tags
FROM
    {table_name}
{where_clauses}\
ORDER BY
    cosine_similarity ASC
LIMIT {limit};
'''

TITLE_WHERE_CLAUSE = 'title % {query_title}  -- Trigram operator (default similarity threshold is 0.3)\n'

PAGE_NUMBERS_WHERE_CLAUSE = 'page_numbers && {query_page_numbers}  -- Overlap operator\n'

TAGS_WHERE_CLAUSE = 'tags && {query_tags}  -- Overlap operator\n'
# ----------------------------------------------------------------------------------------------------------------------
# Generic SQL template for creating a table to hold individual messages from a chatlog and their metadata
CREATE_CHATLOG_TABLE = '''-- Create a table to hold individual messages from a chatlog and their metadata
CREATE TABLE IF NOT EXISTS {table_name} (
    ts TIMESTAMP WITH TIME ZONE PRIMARY KEY,  -- timestamp of the message
    history_key UUID,                         -- history key (unique identifier) of the chatlog this message belongs to
    role INT,                                 -- role of the message
    content TEXT NOT NULL,                    -- text content of the message
    embedding VECTOR({embed_dimension}),      -- embedding vectors (array dimension)
    metadata_JSON JSON                        -- additional metadata of the message
);
'''

INSERT_CHATLOG = '''-- Insert a message into a chatlog
INSERT INTO {table_name} (
    ts,
    history_key,
    role,
    content,
    embedding,
    metadata_JSON
) VALUES ($1, $2, $3, $4, $5, $6);
'''

RETURN_CHATLOG_BY_HISTORY_KEY = '''-- Get entire chatlog of a history key
SELECT
    ts,
    role,
    content,
    metadata_JSON
FROM
    {table_name}
WHERE
    history_key = '{history_key}'
ORDER BY
    ts;
'''

SEMANTIC_QUERY_CHATLOG_TABLE = '''-- Semantic search a chatlog
SELECT
    (embedding <=> '{query_embedding}') AS cosine_similarity,
    ts,
    role,
    content,
    metadata_JSON
FROM
    {table_name}
WHERE
    history_key = '{history_key}'
ORDER BY
    cosine_similarity
LIMIT {limit};
'''
# ======================================================================================================================

ROLE_INTS = {
    'system': 0,
    'user': 1,
    'assistant': 2,
    'tool': 3  # formerly 'function'
}

INT_ROLES = {v: k for k, v in ROLE_INTS.items()}


# Client code could avoid the function call overheads by just doing the dict lookups directly
def role_to_int(role):
    ''' Convert OpenAI style message role from strings to integers for faster DB ops '''
    return ROLE_INTS.get(role, -1)


def int_to_role(role_int):
    ''' Convert OpenAI style message role from integers to strings for faster DB ops '''
    return INT_ROLES.get(role_int, 'unknown')


class PGVectorHelper:
    def __init__(self, embedding_model, table_name: str, apg_conn):
        '''
        Create a PGvector helper from an asyncpg connection

        If you don't yet have a connection, but have all the parameters,
        you can use the PGvectorHelper.from_conn_params() method instead

        Args:
            embedding (SentenceTransformer): SentenceTransformer object of your choice
            https://huggingface.co/sentence-transformers

            table_name: PostgresQL table to store the vector embeddings. Will be checked to restrict to
            alphanumeric characters and underscore

            apg_conn: asyncpg connection to the database
        '''
        if not PREREQS_AVAILABLE:
            raise RuntimeError('pgvector not installed, you can run `pip install pgvector asyncpg`')

        if not table_name.replace('_', '').isalnum():
            raise ValueError('table_name must be alphanumeric, with underscore also allowed')

        # Check if the provided embedding model is a SentenceTransformer
        if (embedding_model.__class__.__name__ == 'SentenceTransformer') and (not None):
            self._embedding_model = embedding_model
            self._embed_dimension = len(self._embedding_model.encode(''))
        elif embedding_model is None:
            self._embedding_model = None
            self._embed_dimension = 0
        else:
            raise ValueError('embedding_model must be a SentenceTransformer object or None')

        self.conn = apg_conn
        self.table_name = table_name

    @classmethod
    async def from_conn_params(cls, embedding_model, table_name, user, 
                               password, db_name, host, port, **conn_params
    ) -> 'PGVectorHelper':
        '''
        Create a PGvector helper from connection parameters

        For details on accepted parameters, See the `pgvector_connection` docstring
            (e.g. run `help(pgvector_connection)`)
        '''
        try:
            conn = await asyncpg.connect(
                host=host,
                port=port,
                user=user,
                password=password,
                database=db_name,
                **conn_params
            )
        except Exception as e:
            # Don't blanket mask the exception. Handle exceptions types in whatever way makes sense
            raise e
        return await cls.from_connection(embedding_model, table_name, conn)

    @classmethod
    async def from_connection(cls, embedding_model, table_name, conn) -> 'PGVectorHelper':
        '''
        Create a PGvector helper from connection parameters

        For details on accepted parameters, See the `pgvector_connection` docstring
            (e.g. run `help(pgvector_connection)`)
        '''
        # Ensure the vector extension is installed
        await conn.execute('CREATE EXTENSION IF NOT EXISTS vector;')
        await register_vector(conn)

        await conn.set_type_codec(  # Register a codec for JSON
            'JSON',
            encoder=json.dumps,
            decoder=json.loads,
            schema='pg_catalog'
        )

        # print('PGvector extension created and loaded.')
        return cls(embedding_model, table_name, conn)

    async def count_items(self) -> int:
        '''
        Count the number of documents in the table

        Returns:
            int: number of documents in the table
        '''
        # Count the number of documents in the table
        count = await self.conn.fetchval(f'SELECT COUNT(*) FROM {self.table_name}')
        return count
    
    async def table_exists(self) -> bool:
        '''
        Check if the table exists

        Returns:
            bool: True if the table exists, False otherwise
        '''
        # Check if the table exists
        table_exists = await self.conn.fetchval(
            CHECK_TABLE_EXISTS,
            self.table_name
        )
        return table_exists

    async def drop_table(self) -> None:
        '''
        Delete the table

        Exercise caution!
        '''
        # Delete the table
        await self.conn.execute(f'DROP TABLE IF EXISTS {self.table_name};')


class DocDB(PGVectorHelper):
    ''' Specialize PGvectorHelper for documents '''
    async def create_table(self) -> None:
        '''
        Create the table to hold embedded documents
        '''
        # Check that the connection is still alive
        # if self.conn.is_closed():
        #     raise ConnectionError('Connection to database is closed')

        # Create the table
        await self.conn.execute(
            CREATE_DOC_TABLE.format(
                table_name=self.table_name,
                embed_dimension=self._embed_dimension)
            )
    
    async def insert(
            self,
            content: str,
            title: str = 'NULL',
            page_numbers: list[int] = [],
            tags: list[str] = []
    ) -> None:
        '''
        Update a table with one embedded document

        Args:
            content (str): text content of the document

            title (str, optional): title of the document

            page_numbers (list[int], optional): page number of the document that the chunk is found in

            tags (list[str], optional): tags associated with the document
        '''
        # Get the embedding of the content as a PGvector compatible list
        content_embedding = self._embedding_model.encode(content)

        await self.conn.execute(
            INSERT_DOCS.format(table_name=self.table_name),
            content_embedding.tolist(),
            content,
            title,
            page_numbers,
            tags
        )

    async def insert_many(
            self,
            content_list: Sequence[tuple[str, str | None,  str | None, list[int], list[str]]]
    ) -> None:
        '''
        Update a table with one or more embedded documents

        Semantically equivalent to multiple insert_doc calls, but uses executemany for efficiency

        Args:
            content_list: List of tuples, each of the form: (content, permission, title, page_numbers, tags)
        '''
        await self.conn.executemany(
            INSERT_DOCS.format(table_name=self.table_name),
            [
                (self._embedding_model.encode(content), content, permission, title, page_numbers, tags)
                for content, permission, title, page_numbers, tags in content_list
            ]
        )

    async def search(
            self,
            query_string: str,
            query_title: str | None = None,
            query_page_numbers: list[int] | None = None,
            query_tags: list[str] | None = None,
            limit: int = 1
    ) -> list[asyncpg.Record]:
        '''
        Similarity search documents using a query string

        Args:
            query_string (str): string to compare against items in the table

            query_title (str, optional): title of the document to compare against items in the table (mildly fuzzy)

            query_page_numbers (list[int], optional): page number of the document that the chunk is found in to compare 
                against items in the table

            query_tags (list[str], optional): tags associated with the document to compare against items in the table

            limit (int, optional): maximum number of results to return (useful for top-k query)
        Returns:
            list[asyncpg.Record]: list of search results
                (asyncpg.Record objects are similar to dicts, but allow for attribute-style access)
        '''
        if not isinstance(limit, int):
            raise TypeError('limit must be an integer')

        # Get the embedding of the query string as a PGvector compatible list
        query_embedding = list(self._embedding_model.encode(query_string))

        # Build where clauses
        if (query_title is None) and (query_page_numbers is None) and (query_tags is None):
            # No where clauses, so don't bother with the WHERE keyword
            where_clauses = ''
        else:  # construct where clauses
            clauses = []
            if query_title is not None:
                clauses.append(TITLE_WHERE_CLAUSE.format(query_title=query_title))
            if query_page_numbers is not None:
                clauses.append(PAGE_NUMBERS_WHERE_CLAUSE.format(query_page_numbers=query_page_numbers))
            if query_tags is not None:
                clauses.append(TAGS_WHERE_CLAUSE.format(query_tags=query_tags))
            clauses = 'AND\n'.join(clauses)  # TODO: move this into the fstring below after py3.12
            where_clauses = f'WHERE\n{clauses}'  # Add the WHERE keyword

        # Search the table
        search_results = await self.conn.fetch(
            QUERY_DOC_TABLE.format(
                table_name=self.table_name,
                query_embedding=query_embedding,
                where_clauses=where_clauses,
                limit=limit
            )
        )
        return search_results
    

class MessageDB(PGVectorHelper):
    ''' Specialize PGvectorHelper for chatlogs '''
    async def create_table(self):
        '''
        Create the table to hold chatlogs
        '''
        # Check that the connection is still alive
        # if self.conn.is_closed():
        #     raise ConnectionError('Connection to database is closed')

        # Create the table
        await self.conn.execute(
            CREATE_CHATLOG_TABLE.format(
                table_name=self.table_name,
                embed_dimension=self._embed_dimension
            )
        )

    async def insert(
            self,
            history_key: UUID,
            role: str,
            content: str,
            timestamp: datetime | None = None,
            metadata: dict | None = None
    ) -> None:
        '''
        Update a table with one embedded document

        Args:
            history_key (str): history key (unique identifier) of the chatlog this message belongs to

            role (str): role of the message (system, user, assistant, tool (formerly 'function'))

            content (str): text content of the message

            timestamp (datetime, optional): timestamp of the message, defaults to current time

            metadata (dict[str, str], optional): additional metadata of the message
        '''
        if not timestamp:
            timestamp = datetime.utcnow().replace(tzinfo=timezone.utc)

        role_int = role_to_int(role)  # Convert from string roles to integer roles

        # Get the embedding of the content as a PGvector compatible list
        content_embedding = self._embedding_model.encode(content)

        await self.conn.execute(
            INSERT_CHATLOG.format(table_name=self.table_name),
            timestamp,
            history_key,
            role_int,
            content,
            content_embedding.tolist(),
            metadata
        )   
    
    async def get_table(
            self,
            history_key: UUID
    ) -> list[asyncpg.Record]:
        '''
        Get the entire chatlog of a history key

        Args:
            history_key (str): history key of the chatlog
        Returns:
            list[asyncpg.Record]: list of chatlog
                (asyncpg.Record objects are similar to dicts, but allow for attribute-style access)
        '''
        # Get the chatlog
        chatlog_records = await self.conn.fetch(
            RETURN_CHATLOG_BY_HISTORY_KEY.format(
                table_name=self.table_name,
                history_key=history_key
            )
        )

        chatlog = [
            {
                'ts': record['ts'],
                'role': int_to_role(record['role']),
                'content': record['content'],
                'metadata': record['metadata_json']
            }
            for record in chatlog_records
        ]

        return chatlog
    
    async def search(
            self,
            history_key: UUID,
            query_string: str,
            limit: int = 1
    ) -> list[asyncpg.Record]:
        '''
        Similarity search documents using a query string

        Args:
            query_string (str): string to compare against items in the table

            k (int, optional): maximum number of results to return (useful for top-k query)
        Returns:
            list[asyncpg.Record]: list of search results
                (asyncpg.Record objects are similar to dicts, but allow for attribute-style access)
        '''
        if not isinstance(limit, int):
            raise TypeError('limit must be an integer')

        # Get the embedding of the query string as a PGvector compatible list
        query_embedding = list(self._embedding_model.encode(query_string))

        # Search the table
        # FIXME: Figure out the SQL injection guard for history_key and limit. 
        # # Not sure SQL Query params is an option here
        records = await self.conn.fetch(
            SEMANTIC_QUERY_CHATLOG_TABLE.format(
                table_name=self.table_name,
                query_embedding=query_embedding,
                history_key=history_key,
                limit=limit
            )
        )

        search_results = [
            {
                'ts': record['ts'],
                'role': int_to_role(record['role']),
                'content': record['content'],
                'metadata': record['metadata_json'],
                'cosine_similarity': record['cosine_similarity']
            }
            for record in records
        ]

        return search_results



'''
pytest test

or

pytest test/test_text_splitter.py
'''
# ruff: noqa: E501

import pytest

from ogbujipt.text_helper import text_splitter

# Also test: chunk_overlap > chunk_size, chunk_overlap == chunk_size, etc.


def test_split_basic(BASIC_EVEN_BLOCK):
    # Need to test chunk_size > total length
    # chunks = text_splitter(BASIC_EVEN_BLOCK, chunk_size=100, chunk_overlap=10)
    chunks = text_splitter(
        BASIC_EVEN_BLOCK, chunk_size=21, chunk_overlap=3, separator='\n')
    # import pprint; pprint.pprint(chunks)  # noqa
    assert len(chunks) == 3
    assert chunks == [
        '0123456789\nabcdefghij', '0123456789\nabcdefghij\nABCDEFGHIJ',
        'abcdefghij\nABCDEFGHIJ\nklmnopqrst'
        ]


# @pytest.mark.skip(reason="Still working on it")
def test_split_poem(COME_THUNDER_POEM):
    # Just use a subset of the poem
    poem_text = COME_THUNDER_POEM.partition('The arrows of God')[0]
    chunks = text_splitter(poem_text, chunk_size=200, chunk_overlap=30)
    # print(chunks[0], '\n---')
    assert chunks[0] == 'Now that the triumphant march has entered the last street corners,\nRemember, O dancers, the thunder among the clouds…\n\nNow that the laughter, broken in two, hangs tremulous between the teeth,\nRemember, O Dancers, the lightning beyond the earth…'
    assert chunks[1] == 'Now that the triumphant march has entered the last street corners,\nRemember, O dancers, the thunder among the clouds…\n\nNow that the laughter, broken in two, hangs tremulous between the teeth,\nRemember, O Dancers, the lightning beyond the earth…\n\nThe smell of blood already floats in the lavender-mist of the afternoon.\nThe death sentence lies in ambush along the corridors of power;\nAnd a great fearful thing already tugs at the cables of the open air,\nA nebula immense and immeasurable, a night of deep waters —\nAn iron dream unnamed and unprintable, a path of stone.'
    assert chunks[2] == 'Now that the laughter, broken in two, hangs tremulous between the teeth,\nRemember, O Dancers, the lightning beyond the earth…\n\nThe smell of blood already floats in the lavender-mist of the afternoon.\nThe death sentence lies in ambush along the corridors of power;\nAnd a great fearful thing already tugs at the cables of the open air,\nA nebula immense and immeasurable, a night of deep waters —\nAn iron dream unnamed and unprintable, a path of stone.\n\nThe drowsy heads of the pods in barren farmlands witness it,\nThe homesteads abandoned in this century’s brush fire witness it:\nThe myriad eyes of deserted corn cobs in burning barns witness it:\nMagic birds with the miracle of lightning flash on their feathers…'
    # import pprint; pprint.pprint(chunks)  # noqa
    assert len(chunks) == 3


def test_separator_not_found(BASIC_EVEN_BLOCK):
    with pytest.warns(UserWarning):
        chunks = text_splitter(
            BASIC_EVEN_BLOCK, chunk_size=21, chunk_overlap=3, separator='!!!')
    assert len(chunks) == 1
    assert len(chunks[0]) == len(BASIC_EVEN_BLOCK)


def test_separator_chunk_size_too_large(BASIC_EVEN_BLOCK):
    # Just use a subset of the poem
    with pytest.warns(UserWarning):
        chunks = text_splitter(
            BASIC_EVEN_BLOCK, chunk_size=100, chunk_overlap=10, separator='!!!')
    assert len(chunks) == 1
    assert len(chunks[0]) == len(BASIC_EVEN_BLOCK)

# def test_zero_overlap(LOREM_IPSUM):
#     chunks = text_splitter(LOREM_IPSUM, chunk_size=100, chunk_overlap=0, separator=' ')
#     assert len(chunks) == 9
#     # chunks should end on the word that takes it over the chunk size
#     assert chunks[0] == 'Lorem ipsum dolor sit amet, consectetur adipiscing elit. Fusce vestibulum nisl eget mauris malesuada,'
#     assert chunks[1] == 'quis facilisis arcu vehicula. Sed consequat, quam ut auctor volutpat, augue ex tincidunt massa, in varius'
#     assert chunks[2] == 'nulla ex vel ipsum. Nullam vitae eros nec ante sagittis luctus. Nullam scelerisque dolor eu orci iaculis,'
#     assert chunks[3] == 'at convallis nulla luctus. Praesent eget ex id arcu facilisis varius vel id neque. Donec non orci eget'
#     assert chunks[4] == 'elit aliquam tempus. Sed at tortor at tortor congue dictum. Nulla varius erat at libero lacinia, id dignissim'
#     assert chunks[5] == 'risus auctor. Ut eu odio vehicula, tincidunt justo ac, viverra erat. Sed nec sem sit amet erat malesuada'
#     assert chunks[6] == 'finibus. Nulla sit amet diam nec dolor tristique dignissim. Sed vehicula, justo nec posuere eleifend,'
#     assert chunks[7] == 'libero ligula interdum neque, at lacinia arcu quam non est. Integer aliquet, erat id dictum euismod, felis'
#     assert chunks[8] == 'libero blandit lorem, nec ullamcorper quam justo at elit.'



from ogbujipt import config

def test_attr_dict():
    d = config.attr_dict({'a': 1, 'b': 2})
    assert d.__class__.__bases__ == (dict,)
    assert d.a == 1
    assert d.b == 2
    assert d['a'] == 1
    assert d['b'] == 2



'''
pytest test

or

pytest test/test_ogbujipt.py
'''
# ruff: noqa: E501

# import pytest

from ogbujipt import oapi_first_choice_text

def test_oapi_first_choice_text(OPENAI_RESPONSE_OBJECT):
    text1 = oapi_first_choice_text(OPENAI_RESPONSE_OBJECT)
    assert text1 == 'The fitness gram pacer test is a multi stage aerobic capacity test that progressively gets more difficult as it continues. The 20 meter pacer test will begin in 30 seconds. Line up at the start.'


'''
pytest test

or

pytest test/embedding/test_pgvector.py

Uses fixtures from ../conftest.py

'''

import pytest
# from unittest.mock import MagicMock, patch
from unittest.mock import MagicMock

import os
from ogbujipt.embedding.pgvector import DocDB
import numpy as np

# XXX: This stanza to go away once mocking is complete - Kai
HOST = os.environ.get('PG_HOST', '0.0.0.0')
DB_NAME = os.environ.get('PG_DATABASE', 'mock_db')
USER = os.environ.get('PG_USER', 'mock_user')
PASSWORD = os.environ.get('PG_PASSWORD', 'mock_password')
PORT = os.environ.get('PG_PORT', 5432)

pacer_copypasta = [  # Demo document
    'The FitnessGram™ Pacer Test is a multistage aerobic capacity test that progressively gets more difficult as it \
        continues.', 
    'The 20 meter pacer test will begin in 30 seconds. Line up at the start.', 
    'The running speed starts slowly, but gets faster each minute after you hear this signal.', 
    '[beep] A single lap should be completed each time you hear this sound.', 
    '[ding] Remember to run in a straight line, and run as long as possible.', 
    'The second time you fail to complete a lap before the sound, your test is over.', 
    'The test will begin on the word start. On your mark, get ready, start.'
]

# XXX: This is to get around the fact that we can't mock the SentenceTransformer class without importing it - Kai
class SentenceTransformer(object):
    def __init__(self, model_name_or_path):
        self.encode = MagicMock()

@pytest.mark.asyncio
async def test_PGv_embed_pacer():
    dummy_model = SentenceTransformer('mock_transformer')
    dummy_model.encode.return_value = np.array([1, 2, 3])
    print(f'EMODEL: {dummy_model}')
    TABLE_NAME = 'embedding_test'
    try:
        vDB = await DocDB.from_conn_params(
            embedding_model=dummy_model,
            table_name=TABLE_NAME,
            db_name=DB_NAME,
            host=HOST,
            port=int(PORT),
            user=USER,
            password=PASSWORD)
    except ConnectionRefusedError:
        pytest.skip("No Postgres instance made available for test. Skipping.", allow_module_level=True)
    
    assert vDB is not None, ConnectionError("Postgres docker instance not available for testing PG code")
    
    # Create tables
    await vDB.drop_table()
    assert await vDB.table_exists() is False, Exception("Table exists before creation")
    await vDB.create_table()
    assert await vDB.table_exists() is True, Exception("Table does not exist after creation")

    # Insert data
    for index, text in enumerate(pacer_copypasta):   # For each line in the copypasta
        await vDB.insert(                            # Insert the line into the table
            content=text,                            # The text to be embedded
            title=f'Pacer Copypasta line {index}',   # Title metadata
            page_numbers=[1, 2, 3],                  # Page number metadata
            tags=['fitness', 'pacer', 'copypasta'],  # Tag metadata
        )

    assert await vDB.count_items() == len(pacer_copypasta), Exception("Not all documents inserted")

    # search table with perfect match
    search_string = '[beep] A single lap should be completed each time you hear this sound.'
    sim_search = await vDB.search(query_string=search_string, limit=3)
    assert sim_search is not None, Exception("No results returned from perfect search")

    await vDB.drop_table()



'''
pytest test

or

pytest test/embedding/test_qdrant.py

Uses the COME_THUNDER_POEM fixture from conftest.py
'''
# import pytest

from ogbujipt.embedding import qdrant
from ogbujipt.embedding.qdrant import qdrant_collection
from ogbujipt.text_helper import text_splitter

qdrant.QDRANT_AVAILABLE = True


class SentenceTransformer:
    """
    Fake class for testing
    """
    def __init__(self, model_name_or_path = None,
                 modules= None,
                 device = None,
                 cache_folder = None,
                 use_auth_token = None
                 ):
        pass

    def encode(self, sentences,
               batch_size: int = 32,
               show_progress_bar= None,
               output_value: str = 'sentence_embedding',
               convert_to_numpy: bool = True,
               convert_to_tensor: bool = False,
               device= None,
               normalize_embeddings: bool = False):
        pass


def test_qdrant_embed_poem(mocker, COME_THUNDER_POEM, CORRECT_STRING):
    # LLM will be downloaded from HuggingFace automatically
    # FIXME: We want to mock this instead, or rather just have a fixture with the results
    # Split the chunks
    chunks = text_splitter(
        COME_THUNDER_POEM, 
        chunk_size=21, 
        chunk_overlap=3, 
        separator='\n'
        )

    collection_name = 'test_collection'

    # TODO: Add more shape to the mocking, to increase the tests's usefulness
    embedding_model = mocker.MagicMock(spec=SentenceTransformer)
    qdrant.models = mocker.MagicMock()
    mock_vparam = object()
    qdrant.models.VectorParams.side_effect = [mock_vparam]
    mocker.patch('ogbujipt.embedding.qdrant.QdrantClient')

    coll = qdrant_collection(name=collection_name, embedding_model=embedding_model)

    # client.count.side_effect = ['count=0']
    coll.db.count.side_effect = lambda collection_name: 'count=0'
    coll.update(chunks)
    coll.db.recreate_collection.assert_called_once_with(
        collection_name='test_collection',
        vectors_config=mock_vparam
        )

    embedding_model.encode.assert_called_with(CORRECT_STRING)

    # Test update/insert into the DB
    mock_pstruct = object()
    qdrant.models.PointStruct.side_effect = lambda id=None, vector=None, payload=None: mock_pstruct

    coll.db.count.reset_mock()
    coll.update(chunks)

    # XXX: Add test with metadata
    coll.db.upsert.assert_called_with(
        collection_name=collection_name,
        points=[mock_pstruct]
        )



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# ogbujipt/demo/alpaca_simple_fix_xml.py
'''
Quick demo—Send an Alpaca-compatible LLM bad XML & ask it to correct

Needs access to an OpenAI-like service. Default assumption is that you
have a self-hosted framework such as llama-cpp-python or text-generation-webui
running. Say it's at my-llm-host:8000, you can do:

python demo/simple_fix_xml.py --apibase=http://localhost:8000

You can alternatively use OpenAI by using the --openai param

Uses a simple completion model, so it probably shouldn't be used with the actual OpenAI service,
now that they've deprecated simple completion in favor of chat completion.
'''

import click

from ogbujipt.llm_wrapper import openai_api
from ogbujipt.prompting.basic import format
from ogbujipt.prompting.model_style import ALPACA_INSTRUCT_INPUT_DELIMITERS


# Command line arguments defined in click decorators
@click.command()
@click.option('--apibase', default='http://127.0.0.1:8000', help='OpenAI API base URL')
@click.option('--llmtemp', default='0.1', type=float, help='LLM temperature')
@click.option('--openai', is_flag=True, default=False, type=bool,
              help='Use live OpenAI API. If you use this option, you must have '
              '"OPENAI_API_KEY" defined in your environmnt')
@click.option('--model', default='', type=str, 
              help='OpenAI model to use (see https://platform.openai.com/docs/models)')
def main(apibase, llmtemp, openai, model):
    # Use OpenAI API if specified, otherwise emulate with supplied host, etc.
    if openai:
        assert not apibase, 'Don\'t use --apibase with --openai'
        oapi = openai_api(model=(model or 'gpt-3.5-turbo'))
    else:
        oapi = openai_api(model=model, api_base=apibase)

    BAD_XML_CODE = '''\
<earth>
<country><b>Russia</country></b>
<capital>Moscow</capital>
</Earth>'''

    # Recommend you use Word Loom for storing, looking up, managing and formatting prompts
    # Left raw for this simple example
    prompt = format(
        'Correct the given XML to make it well-formed',
        contexts= BAD_XML_CODE,
        preamble='You are a helpful assistant, '
        'who answers questions briefly, in 1st grade language',
        delimiters=ALPACA_INSTRUCT_INPUT_DELIMITERS)
    print(prompt, '\n')

    response = oapi(
        prompt=prompt,  # Prompt (Required)
        temperature=llmtemp,  # Temp (Default 1)
        max_tokens=100,  # Max Token length of generated text (Default 16)
        top_p=1,  # AKA nucleus sampling; can increase diversity of the
                  # generated text (Default 1)
        frequency_penalty=0,    # influences the model to favor more or less
                                # frequent tokens (Default 0)
        presence_penalty=1  # influences the model to use new tokens it has
                            # not yet used (Default 0)
        )

    # Response is a json-like object; extract the text
    print('\nFull response data from LLM:\n', response)

    # Response is a json-like object; just get back the text of the response
    response_text = oapi.first_choice_text(response)
    print('\nResponse text from LLM:\n\n', response_text)


# CLI entry point
if __name__ == '__main__':
    main()



# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
# SPDX-License-Identifier: Apache-2.0
# ogbujipt.prompting.model_style

'''
Delimiters for common LLM model prompting styles.

For details see https://github.com/uogbuji/OgbujiPT/wiki/Prompt-templates
'''

import re
from typing import List
from enum import Enum

from ogbujipt.prompting.basic import pdelim, ordering

# __all__ = ['hosted_model_openai']

# Quick cheat sheet:
# Alpaca style = [A…] \n ### Response:
# Alpaca/Instruct style = ### Instruction:\n [A…] ### Response:
# Alpaca/Instruct style, with input = ### Instruction:\n [A…] ### Input:\n ### Response:
# Vicuña style = ### Human:\n [A…] ### Assistant:
# Wizard style = USER:\n [A…] ### ASSISTANT:

UNKNOWN = 'UNKNOWN'


class style(Enum):
    '''
    Marker of different prompring styles. When LLMs are trained, they're tuned
    to expect prompts according to a specific convention, and can become very
    erratic if you use the wrong one
    '''
    ALPACA = 1
    ALPACA_INSTRUCT = 2
    VICUNA = 3
    WIZARD = 4


VICUNA_DELIMITERS = {
    pdelim.PREQUERY: '### USER: ',
    pdelim.POSTQUERY: '\n### ASSISTANT: ',
}

VICUNA_NOHASH_DELIMITERS = {
    pdelim.PREQUERY: 'USER: ',
    pdelim.POSTQUERY: '\n\nASSISTANT:\n',
}

ALPACA_DELIMITERS = {
    pdelim.POSTQUERY: '\n\n### Response:\n',
}

ALPACA_INSTRUCT_DELIMITERS = {  # remove? seems redundant with ALPACA_INSTRUCT_INPUT_DELIMITERS
    pdelim.PREQUERY: '### Instruction:\n',
    pdelim.POSTQUERY: '\n### Response:\n',
}

ALPACA_INSTRUCT_INPUT_DELIMITERS = {
    pdelim.PREQUERY: '### Instruction:\n',
    pdelim.POSTQUERY: '\n\n### Response:\n',
    # Expect a single context item, to treat as the input:
    pdelim.PRE_ALL_CONTEXT: '\n\n### Input:\n',
    pdelim.META_ORDERING: ordering.QUERY_CONTEXT
}

ORCA_DELIMITERS = {
    pdelim.FIXED_PREAMBLE: '### System:\nYou are an AI assistant that follows instruction extremely well. Help as much as you can.',  # noqa E501
    pdelim.PREQUERY: '\n\n### User:\n',
    pdelim.POSTQUERY: '\n\n### Response:\n',
    # Expect a single context item, to treat as the input:
    pdelim.PRE_ALL_CONTEXT: '\n\n### Input:\n',
    pdelim.META_ORDERING: ordering.QUERY_CONTEXT
}

LLAMA_INSTRUCT = {
    pdelim.PRE_PREAMBLE: '### System:\n',
    pdelim.PREQUERY: '\n\n### User:\n',
    pdelim.POSTQUERY: '\n\n### Assistant:\n',
}

AIROBOROS_SUMMARIZATION_DELIMITERS = {
    pdelim.PRE_ALL_CONTEXT: 'User:\nBEGININPUT\nBEGINCONTEXT\n',
    pdelim.PREQUERY: '\nBEGININSTRUCTION\n',
    pdelim.POSTQUERY: '\nENDINSTRUCTION\nASSISTANT: ',
    pdelim.POST_ALL_CONTEXT: '\nENDCONTEXT\nENDINPUT\n'
}

# Closed-context prompting
AIROBOROS_OBEDIENT_DELIMITERS = {
    pdelim.PRECONTEXT: 'BEGININPUT\n',
    pdelim.POSTCONTEXT: '\nENDINPUT',
    pdelim.PRE_CONTEXT_METADATA: 'BEGINCONTEXT',
    pdelim.POST_CONTEXT_METADATA: 'ENDCONTEXT',
    pdelim.PREQUERY: '\nBEGININSTRUCTION\n',
    pdelim.POSTQUERY: '\nENDINSTRUCTION\n'
}

MISTRAL_INSTRUCTION_DELIMITERS = {
    pdelim.PRE_ALL_CONTEXT: '<s>[INST]',
    pdelim.POST_ALL_CONTEXT: '\n[/INST]',
}

CHATML_DELIMITERS = {
    pdelim.PRE_PREAMBLE: '<|im_start|>system\n',
    pdelim.PREQUERY: '<|im_end|>\n<|im_start|>user\n',
    pdelim.POSTQUERY: '<|im_end|>\n<|im_start|>assistant',
}

# If not using the closed-context/obedient prompting, it's just Vicuña style
AIROBOROS_DELIMITERS = VICUNA_DELIMITERS

# XXX: Should this just be a FIXED_PREAMBLE?
AIROBOROS_SUGGESTED_PREAMBLE = 'A chat between a curious user and an artificial intelligence assistant. The assistant gives helpful, detailed, and polite answers to the user\'s questions, and doesn\'t make up answers if it doesn\'t know.'  # noqa E501

AIR_CONOB_INPUT_PRETMPL = '\nBEGINCONTEXT\n{metadata}\nENDCONTEXT\n{context}'

# Define delimeters in OpenAI GPT style
OPENAI_GPT_DELIMITERS = {
    pdelim.PREQUERY: '### USER:\n',
    pdelim.PRECONTEXT: '\n"""\n',
    pdelim.POSTCONTEXT: '\n"""',
    pdelim.POSTQUERY: '\n\n### ASSISTANT:\n',
    pdelim.META_ORDERING: ordering.QUERY_CONTEXT
}

GORILLA_DELIMITERS = {}  # Gorilla doesn't use delimiters, just natural language prompts

# LLAMA_DELIMITERS?, see https://www.reddit.com/r/LocalLLaMA/comments/155po2p/get_llama_2_prompt_format_right/
# or https://huggingface.co/blog/llama2#how-to-prompt-llama-2
'''
<s>[INST] <<SYS>>
{{ system_prompt }}
<</SYS>>

{{ user_message }} [/INST]
'''

# XXX: Rethink in light of word loom
def concat_input_prompts(context_content_pairs):
    '''
    Turn a sequence of context / text pairs into a composite
    Input set for airoboros
    Context is just the key, value pairs in a text string
    '''
    parts = [
        (AIR_CONOB_INPUT_PRETMPL.format(metadata=metas, context=context))
        for metas, context in context_content_pairs
        ]
    return ''.join(parts)


MODEL_NAME_SEARCH_PATTERNS = {
    re.compile('nous-hermes'): [style.ALPACA_INSTRUCT],
    re.compile('wizardlm'): [style.WIZARD],
    # re.compile(''): [style.ALPACA_INSTRUCT],
}


def model_style_from_name(mname: str) -> List[str]:
    '''
    Uses heuristics to figure out the prompting/model style from its name

    >>> from ogbujipt.prompting.model_style import model_style_from_name
    >>> model_style_from_name('path/wizardlm-13b-v1.0-uncensored.ggmlv3.q6_K.bin')
    [<style.WIZARD: 4>]
    >>> from ogbujipt.llm_wrapper import openai_api
    >>> llm_api = openai_api(base_url='http://localhost:8000')
    # Model style hosted via the API
    >>> model_style_from_name(llm_api.hosted_model()[0])
    '''
    # Use the final slash portion, in case it's a full path
    mname = mname.split('/')[-1]
    for pat, styles in MODEL_NAME_SEARCH_PATTERNS.items():
        if pat.search(mname):
            mstyles = styles
            break
    else:
        mstyles = UNKNOWN
    return mstyles



'''
Common tools & resources for test cases
'''
# ruff: noqa: E501

import pytest


@pytest.fixture
def PROMPTING_USER_QUERY():
    return 'what\'s nine plus ten?'


@pytest.fixture
def PROMPTING_CONTEXT():
    return 'The entirety of "Principia Mathematica" by Isaac Newton'

@pytest.fixture
def PROMPTING_CONTEXT_LIST():
    return [
        (
            'author: Isaac Newton\n'
            'title: Principia Mathematica Volume 1\n'
            'date: 1910\n'
            'ISBN-13: 978-0521626063',
            'All mathematical propositions\' are of the form S is P\'; S\' is the subject, P\' the predicate.'
        ),
        (
            'author: Isaac Newton\n'
            'title: Principia Mathematica Volume 2\n'
            'date: 1912\n'
            'ISBN-13: 978-0521626070',
            'It is desirable to find some convenient phrase which shall denote the whole collection of values of a function, or the whole collection of entities of a given type.'
        ),
        (
            'author: Isaac Newton\n'
            'title: Principia Mathematica Volume 3\n'
            'date: 1913\n'
            'ISBN-13: 978-0521626087',
            'In this volume we shall define the arithmetical relations between finite cardinal numbers and also their addition and multiplication.'
        ),
        (
            'author: Isaac Newton\n'
            'title: Principia Mathematica Volume 4\n'
            'date: 1915\n'
            'ISBN-13: 978-6969696969',
            'Finally, we can get onto the real purpose of this work; to explain why 9 + 10 = 21'
        ),
    ]


@pytest.fixture
def BAD_XML_CODE():
    return '''\
<earth>
<country><b>Russia</country></b>
<capital>Moscow</capital>
</Earth>'''


@pytest.fixture
def COME_THUNDER_POEM():
    # One of Christopher Okigbo's greatest poems
    return '''\
Now that the triumphant march has entered the last street corners,
Remember, O dancers, the thunder among the clouds…

Now that the laughter, broken in two, hangs tremulous between the teeth,
Remember, O Dancers, the lightning beyond the earth…

The smell of blood already floats in the lavender-mist of the afternoon.
The death sentence lies in ambush along the corridors of power;
And a great fearful thing already tugs at the cables of the open air,
A nebula immense and immeasurable, a night of deep waters —
An iron dream unnamed and unprintable, a path of stone.

The drowsy heads of the pods in barren farmlands witness it,
The homesteads abandoned in this century’s brush fire witness it:
The myriad eyes of deserted corn cobs in burning barns witness it:
Magic birds with the miracle of lightning flash on their feathers…

The arrows of God tremble at the gates of light,
The drums of curfew pander to a dance of death;

And the secret thing in its heaving
Threatens with iron mask
The last lighted torch of the century…'''


@pytest.fixture
def COME_THUNDER_POEM_CHUNKS():
    return [
        'Now that the triumphant march has entered the last street corners,\nRemember, O dancers, the thunder among the clouds…\n\nNow that the laughter, broken in two, hangs tremulous between the teeth,\nRemember, O Dancers, the lightning beyond the earth…',
        'Now that the triumphant march has entered the last street corners,\nRemember, O dancers, the thunder among the clouds…\n\nNow that the laughter, broken in two, hangs tremulous between the teeth,\nRemember, O Dancers, the lightning beyond the earth…\n\nThe smell of blood already floats in the lavender-mist of the afternoon.\nThe death sentence lies in ambush along the corridors of power;\nAnd a great fearful thing already tugs at the cables of the open air,\nA nebula immense and immeasurable, a night of deep waters —\nAn iron dream unnamed and unprintable, a path of stone.',
        'Now that the laughter, broken in two, hangs tremulous between the teeth,\nRemember, O Dancers, the lightning beyond the earth…\n\nThe smell of blood already floats in the lavender-mist of the afternoon.\nThe death sentence lies in ambush along the corridors of power;\nAnd a great fearful thing already tugs at the cables of the open air,\nA nebula immense and immeasurable, a night of deep waters —\nAn iron dream unnamed and unprintable, a path of stone.\n\nThe drowsy heads of the pods in barren farmlands witness it,\nThe homesteads abandoned in this century’s brush fire witness it:\nThe myriad eyes of deserted corn cobs in burning barns witness it:\nMagic birds with the miracle of lightning flash on their feathers…',
        'The smell of blood already floats in the lavender-mist of the afternoon.\nThe death sentence lies in ambush along the corridors of power;\nAnd a great fearful thing already tugs at the cables of the open air,\nA nebula immense and immeasurable, a night of deep waters —\nAn iron dream unnamed and unprintable, a path of stone.\n\nThe drowsy heads of the pods in barren farmlands witness it,\nThe homesteads abandoned in this century’s brush fire witness it:\nThe myriad eyes of deserted corn cobs in burning barns witness it:\nMagic birds with the miracle of lightning flash on their feathers…\n\nThe arrows of God tremble at the gates of light,\nThe drums of curfew pander to a dance of death;'
        ]


@pytest.fixture
def BASIC_EVEN_BLOCK():
    # Simple example of text with even-length lines
    return '''\
0123456789
abcdefghij
ABCDEFGHIJ
klmnopqrst'''


@pytest.fixture
def LOREM_IPSUM():
    # Lorem ipsum text
    return ('Lorem ipsum dolor sit amet, consectetur adipiscing elit. Fusce vestibulum nisl eget mauris malesuada, '
            'quis facilisis arcu vehicula. Sed consequat, quam ut auctor volutpat, augue ex tincidunt massa, in varius '
            'nulla ex vel ipsum. Nullam vitae eros nec ante sagittis luctus. Nullam scelerisque dolor eu orci iaculis, '
            'at convallis nulla luctus. Praesent eget ex id arcu facilisis varius vel id neque. Donec non orci eget '
            'elit aliquam tempus. Sed at tortor at tortor congue dictum. Nulla varius erat at libero lacinia, id '
            'dignissim risus auctor. Ut eu odio vehicula, tincidunt justo ac, viverra erat. Sed nec sem sit amet erat '
            'malesuada finibus. Nulla sit amet diam nec dolor tristique dignissim. Sed vehicula, justo nec posuere '
            'eleifend, libero ligula interdum neque, at lacinia arcu quam non est. Integer aliquet, erat id dictum '
            'euismod, felis libero blandit lorem, nec ullamcorper quam justo at elit.')


@pytest.fixture
def OPENAI_RESPONSE_OBJECT():
    # Example of OpenAI response object
    # Response generated by prompting with "Recite the Fitness Gram Pacer Test script."
    return {
        "id": "cmpl-abc123",
        "object": "text_completion",
        "created": 1677858242,
        "model": "text-davinci-003",
        "usage": {
            "prompt_tokens": 10,
            "completion_tokens": 128,
            "total_tokens": 138
            },
        'choices': [
            {
                'text': 'The fitness gram pacer test is a multi stage aerobic capacity test that progressively gets more difficult as it continues. The 20 meter pacer test will begin in 30 seconds. Line up at the start.',
                'index': 0,
                "logprobs": None,
                'finish_reason': 'stop'
                },
            {
                'text': 'The running speed starts slowly, but gets faster each minute after you hear this signal. [beep] A single lap should be completed each time you hear this sound. [ding] Remember to run in a straight line, and run as long as possible.',
                'index': 1,
                "logprobs": None,
                'finish_reason': 'stop'
                },
            {
                'text': 'The second time you fail to complete a lap before the sound, your test is over. The test will begin on the word start. On your mark, get ready, start.',
                'index': 2,
                "logprobs": None,
                'finish_reason': 'stop'
                }
            ]
        }


'''
pytest test

or

pytest test/test_vicuña_prompting.py
'''
# import pytest

from ogbujipt.prompting.basic import format
from ogbujipt.prompting.model_style import VICUNA_DELIMITERS


def vicuna_prompt_building(PROMPTING_USER_QUERY):
    EXPECTED_PROMPT = '### USER:\nwhat\'s nine plus ten?\n\n### ASSISTANT:\n'  # noqa: E501

    prompt = format(
        PROMPTING_USER_QUERY,
        delimiters=VICUNA_DELIMITERS
        )
    assert prompt == EXPECTED_PROMPT
